"""
Module related to the parsing of xsdir files. Originally developed at
https://github.com/pyne/pyne
"""
"""
@author: Pyne https://github.com/pyne/pyne

Copyright 2011-2020, the PyNE Development Team. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE PYNE DEVELOPMENT TEAM "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of the stakeholders of the PyNE project or the employers of PyNE developers.
"""

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at:
    https://eupl.eu/1.2/en/
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import os
from typing import List, Tuple


class Xsdir(object):
    """This class stores the information contained in a single MCNP xsdir file.

    Attributes
    ----------
    f : file handle
        The xsdir file.
    filename : str
        Path to the xsdir file.
    directory : str
        Path to the directory containing the xsdir file.
    datapath : str
        The data path specified in the first line of the xsdir file, if it
        exists.
    awr : dict
        Maps material ids to their atomic weight ratios.
    tables : list
        Entries are XsdirTable objects, that appear in the same order as the
        xsdir table lines.

    Notes
    -----
    See MCNP5 User's Guide Volume 3 Appendix K for more information.
    """

    def __init__(self, filename):
        """Parameters
        ----------
        filename : str
            Path to xsdir file.
        """
        self.f = open(filename, 'r')
        self.filename = os.path.abspath(filename)
        self.directory = os.path.dirname(filename)
        self.awr = {}
        self.tables = []

        self.read()

        # It is useful to have a list of the available tables names to be
        # computed only once at initializations
        tablenames = []
        for table in self:
            name = table.name
            zaidname = name[:-4]
            libname = name[-3:]
            tablenames.append((zaidname, libname))
        self.tablenames = tablenames

    def read(self):
        """Populate the Xsdir object by reading the file.
        """
        # Go to beginning of file
        self.f.seek(0)

        # Read first section (DATAPATH)
        line = self.f.readline()
        words = line.split()
        if words:
            if words[0].lower().startswith('datapath'):
                index = line.index('=')
                self.datapath = line[index+1:].strip()

        # Read second section
        line = self.f.readline()
        words = line.split()
        assert len(words) == 3
        assert words[0].lower() == 'atomic'
        assert words[1].lower() == 'weight'
        assert words[2].lower() == 'ratios'

        while True:
            line = self.f.readline()
            words = line.split()

            # Check for end of second section
            if len(words) % 2 != 0 or words[0] == 'directory':
                break

            for zaid, awr in zip(words[::2], words[1::2]):
                self.awr[zaid] = awr

        # Read third section
        while words[0] != 'directory':
            words = self.f.readline().split()

        while True:
            words = self.f.readline().split()
            if not words:
                break

            # Handle continuation lines
            while words[-1] == '+':
                extraWords = self.f.readline().split()
                words = words[:-1] + extraWords  # Correction
            assert len(words) >= 7

            # Create XsdirTable object and add to line
            table = XsdirTable()
            self.tables.append(table)

            # All tables have at least 7 attributes
            table.name = words[0]
            table.awr = float(words[1])
            table.filename = words[2]
            table.access = words[3]
            table.filetype = int(words[4])
            table.address = int(words[5])
            table.tablelength = int(words[6])

            if len(words) > 7:
                table.recordlength = int(words[7])
            if len(words) > 8:
                table.entries = int(words[8])
            if len(words) > 9:
                table.temperature = float(words[9])
            if len(words) > 10:
                table.ptable = (words[10] == 'ptable')

    def find_table(self, name, mode='default'):
        """Find all tables for a given ZAID.

        Modified for JADE, a bug was corrected since table.name do not
        find natural zaids.

        Parameters
        ----------
        name : str
            The ZIAD name.
        mode : str
            if 'default', the default behaviour is used. If 'exact' the zaid is
            searched in that specific library. If 'default-fast' it is run
            in an accellerated loop giving only the libraries names. default is
            'default'.

        Returns
        -------
        tables : list
            All XsdirTable objects for a given ZAID.
        """
        if mode == 'exact':
            # Faster, checks for the exact name
            ans = self._exact_loop(name, self.tablenames)

        elif mode == 'default':
            # Checks all available libraries for the zaid
            tables = []
            for table in self:
                # tablename = table.name.split('.')[0]
                if name == table.name[:-4]:
                    tables.append(table)

            ans = tables

        elif mode == 'default-fast':
            ans = self._all_fast_loop(name, self.tablenames)

        return ans

    @staticmethod
    def _exact_loop(name: str, tablenames: List[Tuple[str, str]]) -> bool:
        for zaidname, libname in tablenames:
            if name == zaidname+'.'+libname:
                return True

        return False

    @staticmethod
    def _all_fast_loop(name: str, tablenames: List[Tuple[str, str]]) -> List[str]:
        libs = []
        for zaidname, libname in tablenames:
            if name == zaidname:
                libs.append(libname)

        return libs

#################  Added by Davide Laghi ###############################  
    def find_zaids(self, lib):
        """Find all zaids for a given library.

        Parameters
        ----------
        lib : str
            The library suffix.

        Returns
        -------
        tables : list
            All XsdirTable objects for a given library.
        """
        
        tables = []
        
        for table in self:
            tablelib = table.name.split('.')[-1]
            if lib == tablelib:
                tables.append(table)

        return tables
############################################################################

    def to_xsdata(self, filename):
        """Writes a Serpent xsdata file for all continuous energy xs tables.

        Parameters
        ----------
        filename : str
            The output filename.

        """
        xsdata = open(filename, 'w')
        for table in self.tables:
            if table.serpent_type == 1:
                xsdata.write(table.to_serpent() + '\n')
        xsdata.close()


    def __iter__(self):
        for table in self.tables:
            yield table

    def nucs(self):
        """Provides a set of the valid nuclide ids for nuclides contained
        in the xsdir.

        Returns
        -------
        valid_nucs : set
            The valid nuclide ids.
        """
        valid_nucs = set(nucname.id(table.name.split('.')[0])
                         for table in self.tables if
                         nucname.isnuclide(table.name.split('.')[0]))
        return valid_nucs


class XsdirTable(object):
    """Stores all information that describes a xsdir table entry, which appears
    as a single line in xsdir file. Attribute names are based off of those
    found in the MCNP5 User's Guide Volume 3, appendix K.

    Attributes
    ----------
    name : str
        The ZAID and library identifier, delimited by a '.'.
    awr : float
        The atomic mass ratio of the nuclide.
    filename : str
        The relative path of the file containing the xs table.
    access : str
       Additional string to specify an access route, such as UNIX directory.
       This entry is typically 0.
    filetype : int
        Describes whether the file contains formated (1) or unformated (2)
        file.
    address : int
        If filetype is 1, address is the line number of the xsdir table. If
        filetype is 2, address is the record number.
    tablelength : int
        Length of the second block of a data table.
    recordlength : int
        Unused for filetype = 1. For filetype = 2, recordlength is the number
        of entires per record times the size (in bytes) of each entry.
    entries : int
        Unused for filetype = 1. For filetype = 2, it is the number of entries
        per record
    temperature : float
        Temperature in MeV for neutron data only.
    ptable : bool
        True if xs table describes continuous energy neutron data with
        unresolved resonance range probability tables.
    """

    def __init__(self):
        self.name = None
        self.awr = None
        self.filename = None
        self.access = None
        self.filetype = None
        self.address = None
        self.tablelength = None
        self.recordlength = None
        self.entries = None
        self.temperature = None
        self.ptable = False

    @property
    def alias(self):
        """Returns the name of the table entry <ZIAD>.<library id>.
        """
        return self.name

    @property
    def serpent_type(self):
        """Converts cross section table type to Serpent format:
            :1: continuous energy (c).
            :2: dosimetry table (y).
            :3: termal (t).
        """
        if self.name.endswith('c'):
            return 1
        elif self.name.endswith('y'):
            return 2
        elif self.name.endswith('t'):
            return 3
        else:
            return None

    @property
    def metastable(self):
        """Returns 1 is xsdir table nuclide is metastable. Returns zero
        otherwise.
        """
        # Only valid for neutron cross-sections
        if not self.name.endswith('c'):
            return

        # Handle special case of Am-242 and Am-242m
        if self.zaid == '95242':
            return 1
        elif self.zaid == '95642':
            return 0

        # All other cases
        A = int(self.name.split('.')[0]) % 1000
        if A > 600:
            return 1
        else:
            return 0

    @property
    def zaid(self):
        """Returns the ZIAD of the nuclide.
        """
        return self.name[:self.name.find('.')]

    def to_serpent(self, directory=''):
        """Converts table to serpent format.

        Parameters
        ----------
        directory : str
            The directory where Serpent data is to be stored.
        """
        # Adjust directory
        if directory:
            if not directory.endswith('/'):
                directory = directory.strip() + '/'

        return "{0} {0} {1} {2} {3} {4} {5:.11e} {6} {7}".format(
            self.name,
            self.serpent_type, self.zaid, 1 if self.metastable else 0,
            self.awr, self.temperature/8.6173423e-11, self.filetype - 1,
            directory + self.filename)


    def __repr__(self):
        return "<XsDirTable: {0}>".format(self.name)


"""This module is related to the parsing of D1S-UNED meshinfo files.

"""
from __future__ import annotations
"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/  
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""
import json
import os
import re
from typing import Optional
from enum import Enum
import numpy as np
from pathlib import Path
from typing import Optional
import pandas as pd


# PARAMETERS
INDEX_KEY_TIME = "Time"
INDEX_KEY_VOXEL = "Voxel"
INDEX_KEY_CELL = "Cell"
INDEX_KEY_MATERIAL = "Material"
COLUMN_KEY_MASS_GRAMS = "Mass [g]"


class CoordinateType(Enum):
    """Only two acceptable coordinate types exist"""

    CART = "cart"
    CYL = "cyl"


class DataMass:

    def __init__(self, dataframe: pd.DataFrame) -> None:
        """
        The dataframe contains the mass information present in the meshinfo
        file. It can be used either to know the masses of the desired
        Voxel-Cell combinations or to relate the material ids with the MCNP
        cell ids and vice-versa.

        Parameters
        ----------
        dataframe : pd.DataFrame
            mass dataframe

        Attributes
        ----------
        dataframe : pd.DataFrame
            mass dataframe

        Examples
        --------
        Example of mass dataframe

        >>> datamass.df
                                Mass [g]
        Voxel Material Cell
        1     0        327608     0.000000
                    327611     0.000000
                        ...
            110      324893  4977.059766
                    324894  1332.655347
                    324896  8458.281897

        """
        self.df: pd.DataFrame = dataframe.sort_index()
        self._check_dataframe_format()

    def _check_dataframe_format(self) -> None:
        try:
            assert self.df.index.names == [
                INDEX_KEY_VOXEL,
                INDEX_KEY_MATERIAL,
                INDEX_KEY_CELL,
            ]
            assert self.df.columns.tolist() == [COLUMN_KEY_MASS_GRAMS]
        except AssertionError as error:
            raise ValueError(
                "The format of the dataframe is not correct. "
                "It has to be a MultiIndex Dataframe with index: "
                f"['{INDEX_KEY_VOXEL}',"
                f" '{INDEX_KEY_MATERIAL}',"
                f" '{INDEX_KEY_CELL}']"
                f" and a column: '{COLUMN_KEY_MASS_GRAMS}'"
            ) from error

    @property
    def materials(self) -> np.ndarray:
        """Returns the unique values of Material index"""
        material_index = self.df.index.names.index(INDEX_KEY_MATERIAL)
        return self.df.index.levels[material_index].values

    def get_filtered_dataframe(
        self,
        voxels: Optional[list] = None,
        materials: Optional[list] = None,
        cells: Optional[list] = None,
    ) -> pd.DataFrame:
        """Returns a copy of the dataframe, if arguments are given, only the
        given values will appear.

        Parameters
        ----------
        voxels : Optional[list], optional
            voxels to filter for, by default None
        materials : Optional[list], optional
            materials to filter for, by default None
        cells : Optional[list], optional
            cells to filter for, by default None

        Returns
        -------
        pd.DataFrame
            filtered dataframe
        """
        mask = np.full(len(self.df), True)
        for key, filter_values in zip(
            [INDEX_KEY_VOXEL, INDEX_KEY_MATERIAL, INDEX_KEY_CELL],
            [voxels, materials, cells],
        ):
            if filter_values is None:
                continue
            mask *= self.df.index.isin(filter_values, level=key)
        return self.df.iloc[mask]

    def get_cells_from_materials(self,
                                 materials: list[int] | None) -> pd.DataFrame:
        """
        Returns an array with all the cell ids that are filled with the
        materials given as argument.

        Parameters
        ----------
        materials : list[int] | None
            materials to be considered for the filter

        Returns
        -------
        pd.DataFrame
            data related to the requested materials
        """
        filtered_dataframe = self.get_filtered_dataframe(materials=materials)
        return filtered_dataframe.index.unique(level=INDEX_KEY_CELL).values

    def save_dataframe_to_hdf5(self, results_path: os.PathLike) -> None:
        """Saves the dataframe for later loading with the load class method.

        Parameters
        ----------
        results_path : os.PathLike
            path to the output folder
        """
        outfile = os.path.join(results_path, "data_mass.hdf5")
        self.df.to_hdf(outfile, key="df")

    @classmethod
    def load(cls, path_to_folder: os.PathLike) -> DataMass:
        """Returns an instance of the class that was saved with the
        :py:meth:`save_dataframe_to_hdf5` method.

        Parameters
        ----------
        path_to_folder : os.PathLike
            path to the folder where the hdf5 file was saved.

        Returns
        -------
        DataMass
            Loaded DataMass object
        """
        infile = os.path.join(path_to_folder, "data_mass.hdf5")
        dataframe = pd.read_hdf(infile)
        return DataMass(dataframe)


class MeshInfo:
    def __init__(self,
                 coordinates: CoordinateType,
                 data_mass: DataMass,
                 vector_i: list[float],
                 vector_j: list[float],
                 vector_k: list[float]) -> None:
        """This class represents the information read in the meshinfo file
        produced by D1SUNED when using a CuV tally. This class represents a
        cartesian mesh or serves as base class for a cylindrical mesh
        (cylindrical meshes have more parameters).

        Parameters
        ----------
        coordinates : CoordinateType
            either cartesian or cylindrical
        data_mass : DataMass
            mass data associated to the mesh
        vector_i : list[float]
            i vector
        vector_j : list[float]
            j vector
        vector_k : list[float]
            k vector

        Attributes
        ----------
        coordinates : CoordinateType
            either cartesian or cylindrical
        data_mass : DataMass
            mass data associated to the mesh
        vector_i : np.array
            i vector
        vector_j : np.array
            j vector
        vector_k : np.array
            k vector
        """

        self.coordinates = coordinates
        self.data_mass = data_mass
        self.vector_i = np.array(vector_i)
        self.vector_j = np.array(vector_j)
        self.vector_k = np.array(vector_k)
        # self.transformation_matrix = None

    def save(self, path_to_results: os.PathLike) -> None:
        """To save the class all the parameters are saved in a JSON file except
        the data_mass which is saved as a HDF5 dataframe.

        Parameters
        ----------
        path_to_results : os.PathLike
            path to the output folder
        """
        json_data = {
            "coordinates": self.coordinates.value,
            "vector_i": self.vector_i.tolist(),
            "vector_j": self.vector_j.tolist(),
            "vector_k": self.vector_k.tolist(),
            # "transformation_matrix": self.transformation_matrix.tolist(),
        }
        outfile = os.path.join(path_to_results, "meshinfo_parameters.json")
        with open(outfile, "w") as infile:
            json.dump(json_data, infile)
        if self.data_mass is not None:
            self.data_mass.save_dataframe_to_hdf5(path_to_results)

    @classmethod
    def load(cls, path_to_folder: os.PathLike) -> MeshInfo | MeshInfoCyl:
        """Given the path to the folder where the class was saved, it reads both
        the JSON file and the data_mass dataframe to recreate the instance of
        the class.

        Parameters
        ----------
        path_to_folder : os.PathLike
            folder where the files to be loaded are stored

        Returns
        -------
        MeshInfo | MeshInfoCyl
            Loaded mesh info
        """
        data_mass = DataMass.load(path_to_folder)
        with open(path_to_folder / "meshinfo_parameters.json", "r") as infile:
            json_data = json.load(infile)
        coordinates = CoordinateType(json_data["coordinates"])
        if coordinates == CoordinateType.CART:
            return MeshInfo(
                coordinates=coordinates,
                data_mass=data_mass,
                vector_i=json_data["vector_i"],
                vector_j=json_data["vector_j"],
                vector_k=json_data["vector_k"],
            )
        return MeshInfoCyl(
            coordinates=coordinates,
            data_mass=data_mass,
            vector_i=json_data["vector_i"],
            vector_j=json_data["vector_j"],
            vector_k=json_data["vector_k"],
            origin=json_data["origin"],
            axis=json_data["axis"],
            vec=json_data["vec"],
        )

    def __eq__(self, __value: MeshInfo | MeshInfoCyl) -> bool:
        if type(__value) not in [MeshInfo, MeshInfoCyl]:
            return False

        try:
            assert self.coordinates == __value.coordinates
            assert (self.vector_i == __value.vector_i).all()
            assert (self.vector_j == __value.vector_j).all()
            assert (self.vector_k == __value.vector_k).all()
            assert self.data_mass.df.equals(
                __value.data_mass.df)
            if type(self) == MeshInfoCyl:
                assert (self.origin == __value.origin).all()
                assert (self.axis == __value.axis).all()
                assert (self.vec == __value.vec).all()

        except AssertionError:
            return False

        return True


class MeshInfoCyl(MeshInfo):

    def __init__(self,
                 coordinates: CoordinateType,
                 data_mass: Optional[DataMass],
                 vector_i: list[float],
                 vector_j: list[float],
                 vector_k: list[float],
                 origin: list[float],
                 axis: list[float],
                 vec: list[float]) -> None:
        """Same as MeshInfo but adding cylindrical coordinates specific
        parameters like origin, axis and vec.

        Parameters
        ----------
        coordinates : CoordinateType
            either cartesian or cylindrical
        data_mass : DataMass
            mass data associated to the mesh
        vector_i : list[float]
            i vector
        vector_j : list[float]
            j vector
        vector_k : list[float]
            k vector
        origin : list[float]
            x, y, z origin of the mesh
        axis : list[float]
            TBD
        vec : list[float]
            TBD

        Attributes
        ----------
        coordinates : CoordinateType
            either cartesian or cylindrical
        data_mass : DataMass
            mass data associated to the mesh
        vector_i : np.array
            i vector
        vector_j : np.array
            j vector
        vector_k : np.array
            k vector
        origin : np.array
            x, y, z origin of the mesh
        axis : np.array
            TBD
        vec : np.array
            TBD
        """
        super().__init__(
            coordinates=coordinates,
            data_mass=data_mass,
            vector_i=vector_i,
            vector_j=vector_j,
            vector_k=vector_k,
        )
        self.origin = np.array(origin)
        self.axis = np.array(axis)
        self.vec = np.array(vec)

    def save(self, path_to_results: os.PathLike) -> None:
        """To save the class all the parameters are saved in a JSON file except
        the data_mass which is saved as a HDF5 dataframe.

        Parameters
        ----------
        path_to_results : os.PathLike
            _description_
        """
        json_data = {
            "coordinates": self.coordinates.value,
            "vector_i": self.vector_i.tolist(),
            "vector_j": self.vector_j.tolist(),
            "vector_k": self.vector_k.tolist(),
            # "transformation_matrix": self.transformation_matrix.tolist(),
            "origin": self.origin.tolist(),
            "axis": self.axis.tolist(),
            "vec": self.vec.tolist(),
        }
        outfile = os.path.join(path_to_results, "meshinfo_parameters.json")
        with open(outfile, "w") as infile:
            json.dump(json_data, infile)
        if self.data_mass is not None:
            self.data_mass.save_dataframe_to_hdf5(path_to_results)


class MeshInfoFile:
    def __init__(self, info: dict[int, MeshInfo | MeshInfoCyl]) -> None:
        """Parser of the meshinfo D1S-UNED file

        Parameters
        ----------
        info : dict[int, MeshInfo  |  MeshInfoCyl]
            contains the info related to the various meshes

        Attributes
        ----------
        info : dict[int, MeshInfo  |  MeshInfoCyl]
            contains the info related to the various meshes

        Examples
        --------
        parse a meshinfo file

        >>> from f4enix.output.meshinfo import MeshInfoFile
        ... mesh_info = MeshInfoFile.from_file('meshinfo')

        """
        self.info = info

    @classmethod
    def from_file(cls, meshinfofile: os.PathLike) -> MeshInfoFile:
        """Reads the meshinfo file and returns a dictionary where the keys are
        the mesh_id and the value are instances of DataMass.

        Parameters
        ----------
        meshinfofile : os.PathLike
            path to the file to be parsed

        Returns
        -------
        MeshInfoFile
            parsed meshinfo file
        """
        info = {}
        with open(meshinfofile, "r", encoding="utf-8") as infile:
            line = infile.readline()
            while line != "":
                if "Mesh tally number:" in line:
                    mesh_id = int(line.split()[-1])
                    mesh_info = _read_individual_mesh(infile)
                    info[mesh_id] = mesh_info
                line = infile.readline()
        return MeshInfoFile(info=info)

    # def get_first_meshinfo(self) -> MeshInfo | MeshInfoCyl:
    #     """Returns the meshinfo of the mesh with the lowest card id."""
    #     keys = list(self.keys())
    #     keys.sort()
    #     return self[keys[0]]


def _read_individual_mesh(infile):
    # skip 3 lines without relevant info
    for _ in range(3):
        infile.readline()
    words = infile.readline().split()
    if words[0] == "origin":
        # the mesh is cylindrical
        mesh_info = _read_header_cyl(words, infile)
    else:
        mesh_info = _read_header_cart(words, infile)
    mesh_info.data_mass = _read_voxels(infile, mesh_info)
    return mesh_info


def _read_header_cyl(words, infile) -> MeshInfoCyl:
    coordinates = CoordinateType.CYL
    origin = [float(x) for x in words[2:5]]
    axis = [float(x) for x in words[7:10]]
    vec = [float(x) for x in words[13:16]]
    words = infile.readline().split()
    vector_i = [float(x) for x in words[2:]]
    words = infile.readline().split()
    vector_j = [float(x) for x in words[2:]]
    # this line is buggy, it says revolutions but the units are in
    # radians, the two numbers may be merged ex: 0.000000006.28318531
    line = infile.readline()
    vector_k = re.findall(r"\d\.\d{8}", line)
    vector_k = [float(x) / (2 * np.pi) for x in vector_k]
    mesh_info = MeshInfoCyl(
        coordinates=coordinates,
        data_mass=None,
        vector_i=vector_i,
        vector_j=vector_j,
        vector_k=vector_k,
        origin=origin,
        axis=axis,
        vec=vec,
    )
    return mesh_info


def _read_header_cart(words, infile) -> MeshInfo:
    coordinates = CoordinateType.CART
    vector_i = [float(x) for x in words[2:]]
    words = infile.readline().split()
    vector_j = [float(x) for x in words[2:]]
    words = infile.readline().split()
    vector_k = [float(x) for x in words[2:]]
    mesh_info = MeshInfo(
        coordinates=coordinates,
        data_mass=None,
        vector_i=vector_i,
        vector_j=vector_j,
        vector_k=vector_k,
    )
    return mesh_info


def _read_voxels(infile, mesh_info) -> DataMass:
    data = {
        INDEX_KEY_VOXEL: [],
        INDEX_KEY_CELL: [],
        INDEX_KEY_MATERIAL: [],
        COLUMN_KEY_MASS_GRAMS: [],
    }
    number_of_voxels = (
        (len(mesh_info.vector_i) - 1)
        * (len(mesh_info.vector_j) - 1)
        * (len(mesh_info.vector_k) - 1)
    )
    for _i in range(number_of_voxels):
        voxel_id, volume, number_of_cells_inside = infile.readline().split()
        voxel_id = int(voxel_id)
        volume = float(volume)
        number_of_cells_inside = int(number_of_cells_inside)
        for _cell in range(number_of_cells_inside):
            cell_id, material_id, cell_mass = _read_cell_part(infile, volume)
            data[INDEX_KEY_VOXEL].append(voxel_id)
            data[INDEX_KEY_CELL].append(cell_id)
            data[INDEX_KEY_MATERIAL].append(material_id)
            data[COLUMN_KEY_MASS_GRAMS].append(cell_mass)
    if mesh_info.coordinates == CoordinateType.CYL:
        # WARNING: we divide the volume by 2 pi due to a bug in the D1S
        # generation of the meshinfo file for cyl coordinates.
        # It provides a volume calculated as if the theta coordinate units
        # were revolutions, but it is actually in radians.
        data[COLUMN_KEY_MASS_GRAMS] = np.array(data[COLUMN_KEY_MASS_GRAMS]) / (
            2 * np.pi
        )
    # the dataframe format has to be that of DataMass
    mass_dataframe = pd.DataFrame(data).set_index(
        [
            INDEX_KEY_VOXEL,
            INDEX_KEY_MATERIAL,
            INDEX_KEY_CELL,
        ]
    )
    data_mass = DataMass(mass_dataframe)
    return data_mass


def _read_cell_part(infile, volume):
    cell_id, density, material, volume_proportion = infile.readline().split()
    cell_id = int(cell_id)
    density = float(density)
    material = int(material)
    volume_proportion = float(volume_proportion)
    cell_volume_inside_voxel = volume * volume_proportion
    cell_mass = cell_volume_inside_voxel * density
    return cell_id, material, cell_mass



"""
This modules is related to handling nuclear data libraries.

It provides support for zaid and elements data and it used for different
operation such as the translation from a library to another of an MCNP input

"""

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/  
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import json
import re
import logging
import pandas as pd
import os
from importlib.resources import files, as_file

from f4enix.input.xsdirpyne import Xsdir
import f4enix.resources as pkg_res

# colors
CRED = '\033[91m'
CEND = '\033[0m'


MSG_DEFLIB = ' The Default library {} was used for zaid {}'


class LibManager:

    def __init__(self, xsdir_file: os.PathLike = None,
                 defaultlib: str = '81c', activationfile: os.PathLike = None,
                 isotopes_file: os.PathLike = None) -> None:
        """
        Object dealing with all complex operations that involves nuclear data

        Parameters
        ----------
        xsdir_file : str or path
            path to the MCNP xsdir reference file. If None (default) the
            default file shipped with the package is used.
        defaultlib : str, optional
            lib suffix to be used as default in translation operations.
            The default is '81c'.
        activationfile : str or path, optional
            path to the config file containing the reactions data for
            activation libraries. The default is None.
        isotopes_file : str or path, optional
            path to the isotopes files. If None (default) the default file
            shipped with the package is used.

        Returns
        -------
        None.

        Attributes
        ----------
        XS: Xsdir
            xsdir file representation
        isotopes: pd.DataFrame
            table containing different data for each isotopes such as formula,
            atomic number, natural abundance, atom mass, etc.
        defaultlib: str
            lib suffix to be used as default in translation operations.
        libraries: list[str]
            list of all the libraries present in the Xsdir file
        reactions: dict[str, pd.DataFrame]
            if an activation file was provided, this stores all the available
            decay reactions in d1s libraries.

        """
        # use both default files
        if xsdir_file is None and isotopes_file is None:
            # get the handlers for the default data files
            resources = files(pkg_res)
            XSDIR_FILE = as_file(resources.joinpath('xsdir.txt'))
            ISOTOPES_FILE = as_file(resources.joinpath('Isotopes.txt'))
            with XSDIR_FILE as xsdir, ISOTOPES_FILE as isotopes:
                abundances = pd.read_csv(isotopes, skiprows=2)
                self.XS = Xsdir(xsdir)

        # use only the default xsdir file
        elif xsdir_file is None:
            # get the handlers for the default data files
            resources = files(pkg_res)
            XSDIR_FILE = as_file(resources.joinpath('xsdir.txt'))
            with XSDIR_FILE as xsdir:
                self.XS = Xsdir(xsdir)
            abundances = pd.read_csv(isotopes_file, skiprows=2)

        # use only the default isotopes file
        elif isotopes_file is None:
            # get the handlers for the default data files
            resources = files(pkg_res)
            ISOTOPES_FILE = as_file(resources.joinpath('Isotopes.txt'))
            with ISOTOPES_FILE as isotopes:
                abundances = pd.read_csv(isotopes, skiprows=2)
            self.XS = Xsdir(xsdir_file)

        # no defaults
        else:
            self.XS = Xsdir(xsdir_file)
            abundances = pd.read_csv(isotopes_file, skiprows=2)

        abundances['idx'] = abundances['idx'].astype(str)
        abundances.set_index('idx', inplace=True)
        self.isotopes = abundances

        self.defaultlib = defaultlib

        # Identify different libraries installed. This is done checking H
        # libraries = self.check4zaid('1001')
        # libraries.extend(self.check4zaid('1000'))  # photons
        libraries = []
        for table in self.XS:
            lib = table.name.split('.')[1]
            if lib not in libraries:
                libraries.append(lib)

        self.libraries = libraries

        # Load the activation reaction data if available
        reactions = {}
        if activationfile is not None:
            file = pd.ExcelFile(activationfile)
        else:
            resources = files(pkg_res)
            with as_file(resources.joinpath('activation_libs.xlsx')) as infile:
                file = pd.ExcelFile(infile)
        for sheet in file.sheet_names:
            # Load the df that also needs to be filled
            reactions[sheet] = file.parse(sheet).ffill()

        # These are needed for faster operations
        newiso = self.isotopes.set_index(['E'])
        newiso = newiso.loc[~newiso.index.duplicated(keep='first')]
        self._newiso_byE = newiso.sort_index()

        newiso = self.isotopes.set_index(['Z'])
        newiso = newiso.loc[~newiso.index.duplicated(keep='first')]
        self._newiso_byZ = newiso.sort_index()

        self.reactions = reactions

    def check4zaid(self, zaid: str) -> list[str]:
        """
        Check which libraries are available for the selected zaid and return it

        Parameters
        ----------
        zaid : str
            zaid string (e.g. 1001).

        Returns
        -------
        libraries : list[str]
            list of libraries (tags) available for the zaid.

        """
        libraries = []
        for libname in self.XS.find_table(zaid, mode='default-fast'):
            libraries.append(libname)

        return libraries

    def convertZaid(self, zaid: str, lib: str
                    ) -> dict[str, tuple[str, float, float]]:
        """
        This methods will convert a zaid into the requested library

        modes:
            - 1to1: there is one to one correspondence for the zaid
            - natural: the zaids will be expanded using the natural abundance
            - absent: the zaid is not available in the library, a default one
              will be used or the natural one if available.

        Parameters
        ----------
        zaid : str
            zaid name (ex. 1001).
        lib : str
            library suffix (ex. 21c).

        Raises
        ------
        ValueError
            if the library is not available in the xsdir file or if there is
            no valid translation for the zaid.

        Returns
        -------
        translation : dict[str, tuple[str, float, float]]
            {zaidname:(lib,nat_abundance,Atomic mass)}.

        """
        # Check if library is available in Xsdir
        if lib not in self.libraries:
            raise ValueError('Library '+lib+' is not available in xsdir file')

        zaidlibs = self.check4zaid(zaid)
        # Natural zaid
        if zaid[-3:] == '000':
            # Check if zaid has natural info
            if self.XS.find_table(zaid+'.'+lib, mode='exact'):
                translation = {zaid: (lib, 1, 1)}  # mass not important

            else:  # Has to be expanded
                translation = {}
                reduced = self.isotopes[self.isotopes['Z'] == int(zaid[:-3])]
                for idx, row in reduced.iterrows():
                    # zaid availability must be checked
                    if self.XS.find_table(idx+'.'+lib, mode='exact'):
                        newlib = lib
                    elif self.XS.find_table(idx+'.'+self.defaultlib,
                                            mode='exact'):
                        logging.warning(
                            MSG_DEFLIB.format(self.defaultlib, zaid))
                        newlib = self.defaultlib
                    else:
                        raise ValueError('No available translation for zaid :' +
                                         zaid+'It is needed for natural zaid expansion.')

                    translation[idx] = (newlib, row['Mean value'],
                                        row['Atomic Mass'])
        # 1to1
        elif lib in zaidlibs:
            translation = {zaid: (lib, 1, 1)}  # mass not important

        # No possible correspondence, natural or default lib has to be used
        else:
            # Check if the natural zaid is available
            natzaid = zaid[:-3]+'000'
            if self.XS.find_table(natzaid+'.'+lib, mode='exact'):
                translation = {natzaid: (lib, 1, 1)}  # mass not important
            # Check if default lib is available
            elif self.XS.find_table(zaid+'.'+self.defaultlib, mode='exact'):
                logging.warning(
                    MSG_DEFLIB.format(self.defaultlib, zaid))
                translation = {zaid: (self.defaultlib, 1, 1)}  # mass not imp
            else:
                # Check if any zaid cross section is available
                libraries = self.check4zaid(zaid)
                # It has to be for the same type of particles
                found = False
                for library in libraries:
                    if library[-1] == lib[-1]:
                        found = True
                # If found no lib is assigned
                if found:
                    translation = {zaid: (None, 1, 1)}  # no masses
                # If no possible translation is found raise error
                else:
                    raise ValueError('No available translation for zaid :' +
                                     zaid)

        return translation

    def get_libzaids(self, lib: str) -> list[str]:
        """
        Given a library, returns all zaids available

        Parameters
        ----------
        lib : str
            suffix of the library.

        Returns
        -------
        zaids : list[str]
            list of zaid names available in the library.

        """
        zaids = []

        for table in self.XS.find_zaids(lib):
            zaid = table.name.split('.')[0]
            if zaid not in zaids:
                zaids.append(zaid)

        return zaids

    def get_zaidname(self, zaid: str) -> tuple[str, str]:
        """
        Given a zaid, its element name and formula are returned. E.g.,
        hydrogen, H1

        Parameters
        ----------
        zaid : str
            zaid number (e.g. 1001 for H1).

        Returns
        -------
        name : str
            element name (e.g. hydrogen).
        formula : str
            isotope name (e.g. H1).

        """
        if type(zaid) == str:
            splitted = zaid.split('.')
            elem = splitted[0][:-3]
            i = int(elem)
            isotope = splitted[0][-3:]

        else:
            i = int(zaid.element)
            isotope = zaid.isotope

        # newiso = self.isotopes.set_index('Z')
        # newiso = newiso.loc[~newiso.index.duplicated(keep='first')]

        name = self._newiso_byZ['Element'].loc[i]
        formula = self._newiso_byZ['E'].loc[i]+'-'+str(int(isotope))

        return name, formula

    def get_zaidnum(self, zaidformula: str) -> str:
        """
        Given a zaid formula return the correct number

        Parameters
        ----------
        zaidformula : str
            name of the zaid, e.g., H1.

        Returns
        -------
        zaidnum : str
            number of the zaid ZZZAA

        """
        # get the table and drop the duplicates

        # newiso = self.isotopes.set_index(['E'])
        # newiso = newiso.loc[~newiso.index.duplicated(keep='first')]

        # split the name
        patnum = re.compile(r'\d+')
        patname = re.compile(r'[a-zA-Z]+')
        try:
            num = patnum.search(zaidformula).group()
            name = patname.search(zaidformula).group()
        except AttributeError:
            raise ValueError('No correspondent zaid found for '+zaidformula)

        atomnumber = self._newiso_byE.loc[name, 'Z']

        zaidnum = "{}{:03d}".format(atomnumber, int(num))

        return zaidnum

    def select_lib(self) -> str:
        """
        Prompt a library input selection with Xsdir availabilty check

        Returns
        -------
        lib : str
            Library to assess.

        """
        error = CRED+'''
 Error: {}
 The selected library is not available.
 '''+CEND
        # Add a counter to avoid falling in an endless loop
        i = 0
        while True:
            i += 1
            lib = input(' Select library (e.g. 31c or 99c-31c): ')
            if lib in self.libraries:
                break

            elif lib[0] == '{':
                libs = json.loads(lib)
                # all libraries should be available
                tocheck = list(libs.values())
                tocheck.extend(list(libs.keys()))
                flag = True
                for val in tocheck:
                    if val not in self.libraries:
                        print(error.format(val))
                        flag = False
                if flag:
                    break

            elif '-' in lib:
                libs = lib.split('-')
                flag = True
                for val in libs:
                    if val not in self.libraries:
                        print(error.format(val))
                        flag = False
                if flag:
                    break

            else:
                print(error.format(lib))

            if i > 20:
                raise ValueError('Too many wrong inputs')
        return lib

    def get_zaid_mass(self, zaid: str) -> float:
        """
        Get the atomic mass of one zaid

        Parameters
        ----------
        zaid : matreader.Zaid
            Zaid to examinate.

        Returns
        -------
        m: float
            atomic mass.

        """
        try:
            m = self.isotopes['Atomic Mass'].loc[zaid.element+zaid.isotope]
        except KeyError:  # It means that it is a natural zaid
            # For a natural zaid the natural abundance mass is used
            df = self.isotopes.reset_index()
            df['Partial mass'] = df['Atomic Mass']*df['Mean value']
            masked = df.set_index('Z').loc[int(zaid.element)]
            m = masked['Partial mass'].sum()

        return float(m)

    def get_reactions(self, lib: str, parent: str) -> list[tuple[str, str]]:
        """
        get the reactions available for a specific zaid and parent nuclide

        Parameters
        ----------
        lib : str
            library suffix as in sheet name of the activation file.
        parent : str
            zaid number of the parent (e.g. 1001).

        Returns
        -------
        reactions : list[tuple[str, str]]
            contains tuple of (MT, daughter).

        """
        reactions = []
        try:
            df = self.reactions[lib].set_index('Parent')
            isotopename, formula = self.get_zaidname(parent)
            formulazaid = formula.replace('-', '')  # eliminate the '-'
            # collect and provide as tuples
            subset = df.loc[formulazaid]
            try:
                for _, row in subset.iterrows():
                    MT = str(int(row['MT']))
                    daughter = row['Daughter']
                    daughter = self.get_zaidnum(daughter)
                    reactions.append((MT, daughter))

            except AttributeError:
                # then is not a DF but a Series
                MT = str(int(subset['MT']))
                daughter = subset['Daughter']
                daughter = self.get_zaidnum(daughter)
                reactions.append((MT, daughter))

        except KeyError:
            # library is not available or parent is not available
            pass

        return reactions
